# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>

"""
Run commands when the Scapy interpreter starts.
"""

import builtins
import code
from io import StringIO
import logging
from queue import Queue
import sys
import threading
import traceback

from scapy.config import conf
from scapy.themes import NoTheme, DefaultTheme, HTMLTheme2, LatexTheme2
from scapy.error import log_scapy, Scapy_Exception
from scapy.utils import tex_escape

from typing import (
    Any,
    Optional,
    TextIO,
    Dict,
    Tuple,
)


#########################
#     Autorun stuff     #
#########################

class StopAutorun(Scapy_Exception):
    code_run = ""


class StopAutorunTimeout(StopAutorun):
    pass


class ScapyAutorunInterpreter(code.InteractiveInterpreter):
    def __init__(self, *args, **kargs):
        # type: (*Any, **Any) -> None
        code.InteractiveInterpreter.__init__(self, *args, **kargs)

    def write(self, data):
        # type: (str) -> None
        pass


def autorun_commands(_cmds, my_globals=None, verb=None):
    # type: (str, Optional[Dict[str, Any]], Optional[int]) -> Any
    sv = conf.verb
    try:
        try:
            if my_globals is None:
                from scapy.main import _scapy_builtins
                my_globals = _scapy_builtins()
            interp = ScapyAutorunInterpreter(locals=my_globals)
            try:
                del builtins.__dict__["scapy_session"]["_"]
            except KeyError:
                pass
            if verb is not None:
                conf.verb = verb
            cmd = ""
            cmds = _cmds.splitlines()
            cmds.append("")  # ensure we finish multi-line commands
            cmds.reverse()
            while True:
                if cmd:
                    sys.stderr.write(sys.__dict__.get("ps2", "... "))
                else:
                    sys.stderr.write(sys.__dict__.get("ps1", ">>> "))

                line = cmds.pop()
                print(line)
                cmd += "\n" + line
                sys.last_value = None
                if interp.runsource(cmd):
                    continue
                if sys.last_value:  # An error occurred
                    traceback.print_exception(sys.last_type,
                                              sys.last_value,
                                              sys.last_traceback.tb_next,
                                              file=sys.stdout)
                    sys.last_value = None
                    return False
                cmd = ""
                if len(cmds) <= 1:
                    break
        except SystemExit:
            pass
    finally:
        conf.verb = sv
    try:
        return builtins.__dict__["scapy_session"]["_"]
    except KeyError:
        return builtins.__dict__.get("_", None)


def autorun_commands_timeout(cmds, timeout=None, **kwargs):
    # type: (str, Optional[int], **Any) -> Any
    """
    Wraps autorun_commands with a timeout that raises StopAutorunTimeout
    on expiration.
    """
    if timeout is None:
        return autorun_commands(cmds, **kwargs)

    q = Queue()  # type: Queue[Any]

    def _runner():
        # type: () -> None
        q.put(autorun_commands(cmds, **kwargs))
    th = threading.Thread(target=_runner)
    th.daemon = True
    th.start()
    th.join(timeout)
    if th.is_alive():
        raise StopAutorunTimeout
    return q.get()


class StringWriter(StringIO):
    """Util to mock sys.stdout and sys.stderr, and
    store their output in a 's' var."""
    def __init__(self, debug=None):
        # type: (Optional[TextIO]) -> None
        self.s = ""
        self.debug = debug
        super().__init__()

    def write(self, x):
        # type: (str) -> int
        # Object can be in the middle of being destroyed.
        if getattr(self, "debug", None) and self.debug:
            self.debug.write(x)
        if getattr(self, "s", None) is not None:
            self.s += x
        return len(x)

    def flush(self):
        # type: () -> None
        if getattr(self, "debug", None) and self.debug:
            self.debug.flush()


def autorun_get_interactive_session(cmds, **kargs):
    # type: (str, **Any) -> Tuple[str, Any]
    """Create an interactive session and execute the
    commands passed as "cmds" and return all output

    :param cmds: a list of commands to run
    :param timeout: timeout in seconds
    :returns: (output, returned) contains both sys.stdout and sys.stderr logs
    """
    sstdout, sstderr, sexcepthook = sys.stdout, sys.stderr, sys.excepthook
    sw = StringWriter()
    h_old = log_scapy.handlers[0]
    log_scapy.removeHandler(h_old)
    log_scapy.addHandler(logging.StreamHandler(stream=sw))
    try:
        try:
            sys.stdout = sys.stderr = sw
            sys.excepthook = sys.__excepthook__
            res = autorun_commands_timeout(cmds, **kargs)
        except StopAutorun as e:
            e.code_run = sw.s
            raise
    finally:
        sys.stdout, sys.stderr, sys.excepthook = sstdout, sstderr, sexcepthook
        log_scapy.removeHandler(log_scapy.handlers[0])
        log_scapy.addHandler(h_old)
    return sw.s, res


def autorun_get_interactive_live_session(cmds, **kargs):
    # type: (str, **Any) -> Tuple[str, Any]
    """Create an interactive session and execute the
    commands passed as "cmds" and return all output

    :param cmds: a list of commands to run
    :param timeout: timeout in seconds
    :returns: (output, returned) contains both sys.stdout and sys.stderr logs
    """
    sstdout, sstderr = sys.stdout, sys.stderr
    sw = StringWriter(debug=sstdout)
    try:
        try:
            sys.stdout = sys.stderr = sw
            res = autorun_commands_timeout(cmds, **kargs)
        except StopAutorun as e:
            e.code_run = sw.s
            raise
    finally:
        sys.stdout, sys.stderr = sstdout, sstderr
    return sw.s, res


def autorun_get_text_interactive_session(cmds, **kargs):
    # type: (str, **Any) -> Tuple[str, Any]
    ct = conf.color_theme
    try:
        conf.color_theme = NoTheme()
        s, res = autorun_get_interactive_session(cmds, **kargs)
    finally:
        conf.color_theme = ct
    return s, res


def autorun_get_live_interactive_session(cmds, **kargs):
    # type: (str, **Any) -> Tuple[str, Any]
    ct = conf.color_theme
    try:
        conf.color_theme = DefaultTheme()
        s, res = autorun_get_interactive_live_session(cmds, **kargs)
    finally:
        conf.color_theme = ct
    return s, res


def autorun_get_ansi_interactive_session(cmds, **kargs):
    # type: (str, **Any) -> Tuple[str, Any]
    ct = conf.color_theme
    try:
        conf.color_theme = DefaultTheme()
        s, res = autorun_get_interactive_session(cmds, **kargs)
    finally:
        conf.color_theme = ct
    return s, res


def autorun_get_html_interactive_session(cmds, **kargs):
    # type: (str, **Any) -> Tuple[str, Any]
    ct = conf.color_theme

    def to_html(s):
        # type: (str) -> str
        return s.replace("<", "&lt;").replace(">", "&gt;").replace("#[#", "<").replace("#]#", ">")  # noqa: E501
    try:
        try:
            conf.color_theme = HTMLTheme2()
            s, res = autorun_get_interactive_session(cmds, **kargs)
        except StopAutorun as e:
            e.code_run = to_html(e.code_run)
            raise
    finally:
        conf.color_theme = ct

    return to_html(s), res


def autorun_get_latex_interactive_session(cmds, **kargs):
    # type: (str, **Any) -> Tuple[str, Any]
    ct = conf.color_theme

    def to_latex(s):
        # type: (str) -> str
        return tex_escape(s).replace("@[@", "{").replace("@]@", "}").replace("@`@", "\\")  # noqa: E501
    try:
        try:
            conf.color_theme = LatexTheme2()
            s, res = autorun_get_interactive_session(cmds, **kargs)
        except StopAutorun as e:
            e.code_run = to_latex(e.code_run)
            raise
    finally:
        conf.color_theme = ct
    return to_latex(s), res



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>

"""Clone of Nmap's first generation OS fingerprinting.

This code works with the first-generation OS detection and
nmap-os-fingerprints, which has been removed from Nmap on November 3,
2007 (https://github.com/nmap/nmap/commit/50c49819), which means it is
outdated.

To get the last published version of this outdated fingerprint
database, you can fetch it from
<https://raw.githubusercontent.com/nmap/nmap/9efe1892/nmap-os-fingerprints>.

"""

import os
import re

from scapy.data import KnowledgeBase
from scapy.config import conf
from scapy.arch import WINDOWS
from scapy.error import warning
from scapy.layers.inet import IP, TCP, UDP, ICMP, UDPerror, IPerror
from scapy.packet import NoPayload, Packet
from scapy.sendrecv import sr
from scapy.compat import plain_str, raw
from scapy.plist import SndRcvList, PacketList

# Typing imports
from typing import (
    Dict,
    List,
    Tuple,
    Optional,
    cast,
    Union,
)

if WINDOWS:
    conf.nmap_base = os.environ["ProgramFiles"] + "\\nmap\\nmap-os-fingerprints"  # noqa: E501
else:
    conf.nmap_base = "/usr/share/nmap/nmap-os-fingerprints"


######################
#  nmap OS fp stuff  #
######################


_NMAP_LINE = re.compile('^([^\\(]*)\\(([^\\)]*)\\)$')


class NmapKnowledgeBase(KnowledgeBase):
    """A KnowledgeBase specialized in Nmap first-generation OS
fingerprints database. Loads from conf.nmap_base when self.filename is
None.

    """

    def lazy_init(self):
        # type: () -> None
        try:
            fdesc = open(conf.nmap_base
                         if self.filename is None else
                         self.filename, "rb")
        except (IOError, TypeError):
            warning("Cannot open nmap database [%s]", self.filename)
            self.filename = None
            return

        self.base = []
        self.base = cast(List[Tuple[str, Dict[str, Dict[str, str]]]], self.base)
        name = None
        sig = {}  # type: Dict[str,Dict[str,str]]
        for line in fdesc:
            str_line = plain_str(line)
            str_line = str_line.split('#', 1)[0].strip()
            if not str_line:
                continue
            if str_line.startswith("Fingerprint "):
                if name is not None:
                    self.base.append((name, sig))
                name = str_line[12:].strip()
                sig = {}
                continue
            if str_line.startswith("Class "):
                continue
            match_line = _NMAP_LINE.search(str_line)
            if match_line is None:
                continue
            test, values = match_line.groups()
            sig[test] = dict(val.split('=', 1) for val in
                             (values.split('%') if values else []))
        if name is not None:
            self.base.append((name, sig))
        fdesc.close()

    def get_base(self):
        # type: () -> List[Tuple[str, Dict]]
        return cast(List[Tuple[str, Dict]], super(NmapKnowledgeBase, self).get_base())


conf.nmap_kdb = NmapKnowledgeBase(None)
conf.nmap_kdb = cast(NmapKnowledgeBase, conf.nmap_kdb)


def nmap_tcppacket_sig(pkt):
    # type: (Optional[Packet]) -> Dict
    res = {}
    if pkt is not None:
        res["DF"] = "Y" if pkt.flags.DF else "N"
        res["W"] = "%X" % pkt.window
        res["ACK"] = "S++" if pkt.ack == 2 else "S" if pkt.ack == 1 else "O"
        res["Flags"] = str(pkt[TCP].flags)[::-1]
        res["Ops"] = "".join(x[0][0] for x in pkt[TCP].options)
    else:
        res["Resp"] = "N"
    return res


def nmap_udppacket_sig(snd, rcv):
    # type: (SndRcvList, PacketList) -> Dict
    res = {}
    if rcv is None:
        res["Resp"] = "N"
    else:
        res["DF"] = "Y" if rcv.flags.DF else "N"
        res["TOS"] = "%X" % rcv.tos
        res["IPLEN"] = "%X" % rcv.len
        res["RIPTL"] = "%X" % rcv.payload.payload.len
        res["RID"] = "E" if snd.id == rcv[IPerror].id else "F"
        res["RIPCK"] = "E" if snd.chksum == rcv[IPerror].chksum else (
            "0" if rcv[IPerror].chksum == 0 else "F"
        )
        res["UCK"] = "E" if snd.payload.chksum == rcv[UDPerror].chksum else (
            "0" if rcv[UDPerror].chksum == 0 else "F"
        )
        res["ULEN"] = "%X" % rcv[UDPerror].len
        res["DAT"] = "E" if (
            isinstance(rcv[UDPerror].payload, NoPayload) or
            raw(rcv[UDPerror].payload) == raw(snd[UDP].payload)
        ) else "F"
    return res


def nmap_match_one_sig(seen, ref):
    # type: (Dict, Dict) -> float
    cnt = sum(val in ref.get(key, "").split("|") for key, val in seen.items())
    if cnt == 0 and seen.get("Resp") == "N":
        return 0.7
    return float(cnt) / len(seen)


def nmap_sig(target, oport=80, cport=81, ucport=1):
    # type: (str, int, int, int) -> Dict
    res = {}

    tcpopt = [("WScale", 10),
              ("NOP", None),
              ("MSS", 256),
              ("Timestamp", (123, 0))]
    tests = [
        IP(dst=target, id=1) /
        TCP(seq=1, sport=5001 + i, dport=oport if i < 4 else cport,
            options=tcpopt, flags=flags)
        for i, flags in enumerate(["CS", "", "SFUP", "A", "S", "A", "FPU"])
    ]
    tests.append(IP(dst=target) / UDP(sport=5008, dport=ucport) / (300 * "i"))

    ans, unans = sr(tests, timeout=2)
    ans.extend((x, None) for x in unans)

    for snd, rcv in ans:
        if snd.sport == 5008:
            res["PU"] = (snd, rcv)
        else:
            test = "T%i" % (snd.sport - 5000)
            if rcv is not None and ICMP in rcv:
                warning("Test %s answered by an ICMP", test)
                rcv = None  # type: ignore
            res[test] = rcv

    return nmap_probes2sig(res)


def nmap_probes2sig(tests):
    # type: (Dict) -> Dict
    tests = tests.copy()
    res = {}
    if "PU" in tests:
        res["PU"] = nmap_udppacket_sig(*tests["PU"])
        del tests["PU"]
    for k in tests:
        res[k] = nmap_tcppacket_sig(tests[k])
    return res


def nmap_search(sigs):
    # type: (Dict) -> Tuple[Union[int, float], List]
    guess = 0, []  # type: Tuple[Union[int, float], List]
    conf.nmap_kdb = cast(NmapKnowledgeBase, conf.nmap_kdb)
    for osval, fprint in conf.nmap_kdb.get_base():
        score = 0.0
        for test, values in fprint.items():
            if test in sigs:
                score += nmap_match_one_sig(sigs[test], values)
        score /= len(sigs)
        if score > guess[0]:
            guess = score, [osval]
        elif score == guess[0]:
            guess[1].append(osval)
    return guess


@conf.commands.register
def nmap_fp(target, oport=80, cport=81):
    # type: (str, int, int) -> Tuple[Union[int, float], List]
    """nmap fingerprinting
nmap_fp(target, [oport=80,] [cport=81,]) -> list of best guesses with accuracy
"""
    sigs = nmap_sig(target, oport, cport)
    return nmap_search(sigs)


@conf.commands.register
def nmap_sig2txt(sig):
    # type: (Dict) -> str
    torder = ["TSeq", "T1", "T2", "T3", "T4", "T5", "T6", "T7", "PU"]
    korder = ["Class", "gcd", "SI", "IPID", "TS",
              "Resp", "DF", "W", "ACK", "Flags", "Ops",
              "TOS", "IPLEN", "RIPTL", "RID", "RIPCK", "UCK", "ULEN", "DAT"]
    txt = []
    for i in sig:
        if i not in torder:
            torder.append(i)
    for test in torder:
        testsig = sig.get(test)
        if testsig is None:
            continue
        txt.append("%s(%s)" % (test, "%".join(
            "%s=%s" % (key, testsig[key]) for key in korder if key in testsig
        )))
    return "\n".join(txt)



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

"""
DCE/RPC client as per [MS-RPCE]
"""

import uuid
import socket

from scapy.config import conf
from scapy.error import log_runtime

from scapy.layers.dcerpc import (
    _DCE_RPC_ERROR_CODES,
    ComInterface,
    CommonAuthVerifier,
    DCE_C_AUTHN_LEVEL,
    DCERPC_Transport,
    DceRpc5,
    DceRpc5AbstractSyntax,
    DceRpc5AlterContext,
    DceRpc5AlterContextResp,
    DceRpc5Auth3,
    DceRpc5Bind,
    DceRpc5BindAck,
    DceRpc5BindNak,
    DceRpc5Context,
    DceRpc5Fault,
    DceRpc5Request,
    DceRpc5Response,
    DceRpc5TransferSyntax,
    DceRpcInterface,
    DceRpcSecVT,
    DceRpcSecVTCommand,
    DceRpcSecVTPcontext,
    DceRpcSession,
    DceRpcSocket,
    find_dcerpc_interface,
    NDRContextHandle,
    NDRPointer,
    RPC_C_IMP_LEVEL,
)
from scapy.layers.gssapi import (
    SSP,
    GSS_S_FAILURE,
    GSS_S_COMPLETE,
    GSS_S_CONTINUE_NEEDED,
    GSS_C_FLAGS,
)
from scapy.layers.smb2 import STATUS_ERREF
from scapy.layers.smbclient import (
    SMB_RPC_SOCKET,
)

# RPC
from scapy.layers.msrpce.ept import (
    ept_map_Request,
    ept_map_Response,
    twr_p_t,
    protocol_tower_t,
    prot_and_addr_t,
    UUID,
)

# Typing
from typing import (
    Optional,
    Union,
)


class DCERPC_Client(object):
    """
    A basic DCE/RPC client

    :param transport: the transport to use.
    :param ndr64: should ask for NDR64 when binding (default conf.ndr64)
    :param ndrendian: the endianness to use (default little)
    :param verb: enable verbose logging (default True)
    :param auth_level: the DCE_C_AUTHN_LEVEL to use
    :param impersonation_type: the RPC_C_IMP_LEVEL to use
    """

    def __init__(
        self,
        transport: DCERPC_Transport,
        ndr64: Optional[bool] = None,
        ndrendian: str = "little",
        verb: bool = True,
        auth_level: Optional[DCE_C_AUTHN_LEVEL] = None,
        impersonation_type: RPC_C_IMP_LEVEL = RPC_C_IMP_LEVEL.DEFAULT,
        **kwargs,
    ):
        self.sock = None
        self.transport = transport
        assert isinstance(
            transport, DCERPC_Transport
        ), "transport must be from DCERPC_Transport"

        # Counters
        self.call_id = 0
        self.next_cont_id = 0  # next available context id
        self.next_auth_contex_id = 0  # next available auth context id

        # Session parameters
        if ndr64 is None:
            ndr64 = conf.ndr64
        self.ndr64: bool = ndr64
        self.ndrendian = ndrendian
        self.verb = verb
        self.host: str = None
        self.port: int = -1
        self.ssp = kwargs.pop("ssp", None)  # type: SSP
        self.sspcontext = None
        if auth_level is not None:
            self.auth_level = auth_level
        elif self.ssp is not None:
            self.auth_level = DCE_C_AUTHN_LEVEL.CONNECT
        else:
            self.auth_level = DCE_C_AUTHN_LEVEL.NONE
        if impersonation_type == RPC_C_IMP_LEVEL.DEFAULT:
            # Same default as windows
            impersonation_type = RPC_C_IMP_LEVEL.IDENTIFY
        self.impersonation_type = impersonation_type
        self._first_time_on_interface = True
        self.contexts = {}
        self.dcesockargs = kwargs
        self.dcesockargs["transport"] = self.transport

    @classmethod
    def from_smblink(cls, smbcli, smb_kwargs={}, **kwargs):
        """
        Build a DCERPC_Client from a SMB_Client.smblink directly
        """
        client = DCERPC_Client(DCERPC_Transport.NCACN_NP, **kwargs)
        sock = client.smbrpcsock = SMB_RPC_SOCKET(smbcli, **smb_kwargs)
        client.sock = DceRpcSocket(
            sock,
            DceRpc5,
            ssp=client.ssp,
            auth_level=client.auth_level,
            **client.dcesockargs,
        )
        return client

    @property
    def session(self) -> DceRpcSession:
        return self.sock.session

    def connect(
        self,
        host,
        endpoint: Union[int, str] = None,
        port: Optional[int] = None,
        interface=None,
        timeout=5,
        smb_kwargs={},
    ):
        """
        Initiate a connection.

        :param host: the host to connect to
        :param endpoint: (optional) the port/smb pipe to connect to
        :param interface: (optional) if endpoint isn't provided, uses the endpoint
            mapper to find the appropriate endpoint for that interface.
        :param timeout: (optional) the connection timeout (default 5)
        :param port: (optional) the port to connect to. (useful for SMB)
        """
        if endpoint is None and interface is not None:
            # Figure out the endpoint using the endpoint mapper

            if self.transport == DCERPC_Transport.NCACN_IP_TCP and port is None:
                # IP/TCP
                # ask the endpoint mapper (port 135) for the IP:PORT
                endpoints = get_endpoint(
                    host,
                    interface,
                    ndrendian=self.ndrendian,
                    verb=self.verb,
                )
                if endpoints:
                    _, endpoint = endpoints[0]
                else:
                    raise ValueError(
                        "Could not find an available endpoint for that interface !"
                    )
            elif self.transport == DCERPC_Transport.NCACN_NP:
                # SMB
                # ask the endpoint mapper (over SMB) for the namedpipe
                endpoints = get_endpoint(
                    host,
                    interface,
                    transport=self.transport,
                    ndrendian=self.ndrendian,
                    verb=self.verb,
                    smb_kwargs=smb_kwargs,
                )
                if endpoints:
                    endpoint = endpoints[0].lstrip("\\pipe\\")
                else:
                    return

        # Assign the default port if no port is provided
        if port is None:
            if self.transport == DCERPC_Transport.NCACN_IP_TCP:  # IP/TCP
                port = endpoint or 135
            elif self.transport == DCERPC_Transport.NCACN_NP:  # SMB
                port = 445
            else:
                raise ValueError(
                    "Can't guess the port for transport: %s" % self.transport
                )

        # Start socket and connect
        self.host = host
        self.port = port
        sock = socket.socket()
        sock.settimeout(timeout)
        if self.verb:
            print(
                "\u2503 Connecting to %s on port %s via %s..."
                % (host, port, repr(self.transport))
            )
        sock.connect((host, port))
        if self.verb:
            print(
                conf.color_theme.green(
                    "\u2514 Connected from %s" % repr(sock.getsockname())
                )
            )

        if self.transport == DCERPC_Transport.NCACN_NP:  # SMB
            # If the endpoint is provided, connect to it.
            if endpoint is not None:
                self.open_smbpipe(endpoint)

            # We pack the socket into a SMB_RPC_SOCKET
            sock = self.smbrpcsock = SMB_RPC_SOCKET.from_tcpsock(
                sock, ssp=self.ssp, **smb_kwargs
            )
            self.sock = DceRpcSocket(sock, DceRpc5, **self.dcesockargs)
        elif self.transport == DCERPC_Transport.NCACN_IP_TCP:
            self.sock = DceRpcSocket(
                sock,
                DceRpc5,
                ssp=self.ssp,
                auth_level=self.auth_level,
                **self.dcesockargs,
            )

    def close(self):
        """
        Close the DCE/RPC client.
        """
        if self.verb:
            print("X Connection closed\n")
        self.sock.close()

    def sr1(self, pkt, **kwargs):
        """
        Send/Receive a DCE/RPC message.

        The DCE/RPC header is added automatically.
        """
        self.call_id += 1
        pkt = (
            DceRpc5(
                call_id=self.call_id,
                pfc_flags="PFC_FIRST_FRAG+PFC_LAST_FRAG",
                endian=self.ndrendian,
                auth_verifier=kwargs.pop("auth_verifier", None),
                vt_trailer=kwargs.pop("vt_trailer", None),
            )
            / pkt
        )
        if "pfc_flags" in kwargs:
            pkt.pfc_flags = kwargs.pop("pfc_flags")
        if "objectuuid" in kwargs:
            pkt.pfc_flags += "PFC_OBJECT_UUID"
            pkt.object = kwargs.pop("objectuuid")
        return self.sock.sr1(pkt, verbose=0, **kwargs)

    def send(self, pkt, **kwargs):
        """
        Send a DCE/RPC message.

        The DCE/RPC header is added automatically.
        """
        self.call_id += 1
        pkt = (
            DceRpc5(
                call_id=self.call_id,
                pfc_flags="PFC_FIRST_FRAG+PFC_LAST_FRAG",
                endian=self.ndrendian,
                auth_verifier=kwargs.pop("auth_verifier", None),
                vt_trailer=kwargs.pop("vt_trailer", None),
            )
            / pkt
        )
        if "pfc_flags" in kwargs:
            pkt.pfc_flags = kwargs.pop("pfc_flags")
        if "objectuuid" in kwargs:
            pkt.pfc_flags += "PFC_OBJECT_UUID"
            pkt.object = kwargs.pop("objectuuid")
        return self.sock.send(pkt, **kwargs)

    def sr1_req(self, pkt, **kwargs):
        """
        Send/Receive a DCE/RPC request.

        :param pkt: the inner DCE/RPC message, without any header.
        """
        if self.verb:
            if "objectuuid" in kwargs:
                # COM
                print(
                    conf.color_theme.opening(
                        ">> REQUEST (COM): %s" % pkt.payload.__class__.__name__
                    )
                )
            else:
                print(
                    conf.color_theme.opening(">> REQUEST: %s" % pkt.__class__.__name__)
                )
        # Add sectrailer if first time talking on this interface
        vt_trailer = b""
        if (
            self._first_time_on_interface
            and self.transport != DCERPC_Transport.NCACN_NP
        ):
            # In the first request after a bind, Windows sends a trailer to verify
            # that the negotiated transfer/interface wasn't altered.
            self._first_time_on_interface = False
            vt_trailer = DceRpcSecVT(
                commands=[
                    DceRpcSecVTCommand(SEC_VT_COMMAND_END=1)
                    / DceRpcSecVTPcontext(
                        InterfaceId=self.session.rpc_bind_interface.uuid,
                        TransferSyntax="NDR64" if self.ndr64 else "NDR 2.0",
                        TransferVersion=1 if self.ndr64 else 2,
                    )
                ]
            )

        # Optional: force opnum
        opnum = {}
        if "opnum" in kwargs:
            opnum["opnum"] = kwargs.pop("opnum")

        # Send/receive
        resp = self.sr1(
            DceRpc5Request(
                cont_id=self.session.cont_id,
                alloc_hint=len(pkt) + len(vt_trailer),
                **opnum,
            )
            / pkt,
            vt_trailer=vt_trailer,
            **kwargs,
        )

        # Parse result
        result = None
        if DceRpc5Response in resp:
            if self.verb:
                if "objectuuid" in kwargs:
                    # COM
                    print(
                        conf.color_theme.success(
                            "<< RESPONSE (COM): %s"
                            % (resp[DceRpc5Response].payload.payload.__class__.__name__)
                        )
                    )
                else:
                    print(
                        conf.color_theme.success(
                            "<< RESPONSE: %s"
                            % (resp[DceRpc5Response].payload.__class__.__name__)
                        )
                    )
            result = resp[DceRpc5Response].payload
        elif DceRpc5Fault in resp:
            if self.verb:
                print(conf.color_theme.success("<< FAULT"))
                # If [MS-EERR] is loaded, show the extended info
                if resp[DceRpc5Fault].payload and not isinstance(
                    resp[DceRpc5Fault].payload, conf.raw_layer
                ):
                    resp[DceRpc5Fault].payload.show()
            result = resp
        if self.verb and getattr(resp, "status", 0) != 0:
            if resp.status in _DCE_RPC_ERROR_CODES:
                print(conf.color_theme.fail(f"! {_DCE_RPC_ERROR_CODES[resp.status]}"))
            elif resp.status in STATUS_ERREF:
                print(conf.color_theme.fail(f"! {STATUS_ERREF[resp.status]}"))
            else:
                print(conf.color_theme.fail("! Failure"))
                resp.show()
        return result

    def _get_bind_context(self, interface):
        """
        Internal: get the bind DCE/RPC context.
        """
        if interface in self.contexts:
            # We have already found acceptable contexts for this interface,
            # reuse that.
            return self.contexts[interface]

        # NDR 2.0
        contexts = [
            DceRpc5Context(
                cont_id=self.next_cont_id,
                abstract_syntax=DceRpc5AbstractSyntax(
                    if_uuid=interface.uuid,
                    if_version=interface.if_version,
                ),
                transfer_syntaxes=[
                    DceRpc5TransferSyntax(
                        # NDR 2.0 32-bit
                        if_uuid="NDR 2.0",
                        if_version=2,
                    )
                ],
            ),
        ]
        self.next_cont_id += 1

        # NDR64
        if self.ndr64:
            contexts.append(
                DceRpc5Context(
                    cont_id=self.next_cont_id,
                    abstract_syntax=DceRpc5AbstractSyntax(
                        if_uuid=interface.uuid,
                        if_version=interface.if_version,
                    ),
                    transfer_syntaxes=[
                        DceRpc5TransferSyntax(
                            # NDR64
                            if_uuid="NDR64",
                            if_version=1,
                        )
                    ],
                )
            )
            self.next_cont_id += 1

        # BindTimeFeatureNegotiationBitmask
        contexts.append(
            DceRpc5Context(
                cont_id=self.next_cont_id,
                abstract_syntax=DceRpc5AbstractSyntax(
                    if_uuid=interface.uuid,
                    if_version=interface.if_version,
                ),
                transfer_syntaxes=[
                    DceRpc5TransferSyntax(
                        if_uuid=uuid.UUID("6cb71c2c-9812-4540-0300-000000000000"),
                        if_version=1,
                    )
                ],
            )
        )
        self.next_cont_id += 1

        # Store contexts for this interface
        self.contexts[interface] = contexts

        return contexts

    def _check_bind_context(self, interface, contexts) -> bool:
        """
        Internal: check the answer DCE/RPC bind context, and update them.
        """
        for i, ctx in enumerate(contexts):
            if ctx.result == 0:
                # Context was accepted. Remove all others from cache
                self.contexts[interface] = [self.contexts[interface][i]]
                return True

        return False

    def _bind(
        self, interface: Union[DceRpcInterface, ComInterface], reqcls, respcls
    ) -> bool:
        """
        Internal: used to send a bind/alter request
        """
        # Build a security context: [MS-RPCE] 3.3.1.5.2
        if self.verb:
            print(
                conf.color_theme.opening(
                    ">> %s on %s" % (reqcls.__name__, interface)
                    + (" (with %s)" % self.ssp.__class__.__name__ if self.ssp else "")
                )
            )

        # Do we need an authenticated bind
        if not self.ssp or (
            self.sspcontext is not None
            or self.transport == DCERPC_Transport.NCACN_NP
            and self.auth_level < DCE_C_AUTHN_LEVEL.PKT_INTEGRITY
        ):
            # NCACN_NP = SMB without INTEGRITY/PRIVACY does not bind the RPC securely,
            # again as it has already authenticated during the SMB Session Setup
            resp = self.sr1(
                reqcls(context_elem=self._get_bind_context(interface)),
                auth_verifier=None,
            )
            status = GSS_S_COMPLETE
        else:
            # Perform authentication
            self.sspcontext, token, status = self.ssp.GSS_Init_sec_context(
                self.sspcontext,
                req_flags=(
                    # SSPs need to be instantiated with some special flags
                    # for DCE/RPC usages.
                    GSS_C_FLAGS.GSS_C_DCE_STYLE
                    | GSS_C_FLAGS.GSS_C_REPLAY_FLAG
                    | GSS_C_FLAGS.GSS_C_SEQUENCE_FLAG
                    | GSS_C_FLAGS.GSS_C_MUTUAL_FLAG
                    | (
                        GSS_C_FLAGS.GSS_C_INTEG_FLAG
                        if self.auth_level >= DCE_C_AUTHN_LEVEL.PKT_INTEGRITY
                        else 0
                    )
                    | (
                        GSS_C_FLAGS.GSS_C_CONF_FLAG
                        if self.auth_level >= DCE_C_AUTHN_LEVEL.PKT_PRIVACY
                        else 0
                    )
                    | (
                        GSS_C_FLAGS.GSS_C_IDENTIFY_FLAG
                        if self.impersonation_type <= RPC_C_IMP_LEVEL.IDENTIFY
                        else 0
                    )
                    | (
                        GSS_C_FLAGS.GSS_C_DELEG_FLAG
                        if self.impersonation_type == RPC_C_IMP_LEVEL.DELEGATE
                        else 0
                    )
                ),
                target_name="host/" + self.host,
            )

            if status not in [GSS_S_CONTINUE_NEEDED, GSS_S_COMPLETE]:
                # Authentication failed.
                self.sspcontext.clifailure()
                return False

            resp = self.sr1(
                reqcls(context_elem=self._get_bind_context(interface)),
                auth_verifier=(
                    None
                    if not self.sspcontext
                    else CommonAuthVerifier(
                        auth_type=self.ssp.auth_type,
                        auth_level=self.auth_level,
                        auth_context_id=self.session.auth_context_id,
                        auth_value=token,
                    )
                ),
                pfc_flags=(
                    "PFC_FIRST_FRAG+PFC_LAST_FRAG"
                    + (
                        # If the SSP supports "Header Signing", advertise it
                        "+PFC_SUPPORT_HEADER_SIGN"
                        if self.ssp is not None and self.session.support_header_signing
                        else ""
                    )
                ),
            )

            # Check that the answer looks valid and contexts were accepted
            if respcls not in resp or not self._check_bind_context(
                interface, resp.results
            ):
                token = None
                status = GSS_S_FAILURE
            else:
                # Call the underlying SSP
                self.sspcontext, token, status = self.ssp.GSS_Init_sec_context(
                    self.sspcontext,
                    input_token=resp.auth_verifier.auth_value,
                    target_name="host/" + self.host,
                )

            if status in [GSS_S_CONTINUE_NEEDED, GSS_S_COMPLETE]:
                # Authentication should continue, in two ways:
                # - through DceRpc5Auth3 (e.g. NTLM)
                # - through DceRpc5AlterContext (e.g. Kerberos)
                if token and self.ssp.LegsAmount(self.sspcontext) % 2 == 1:
                    # AUTH 3 for certain SSPs (e.g. NTLM)
                    # "The server MUST NOT respond to an rpc_auth_3 PDU"
                    self.send(
                        DceRpc5Auth3(),
                        auth_verifier=CommonAuthVerifier(
                            auth_type=self.ssp.auth_type,
                            auth_level=self.auth_level,
                            auth_context_id=self.session.auth_context_id,
                            auth_value=token,
                        ),
                    )
                    status = GSS_S_COMPLETE
                else:
                    while token:
                        respcls = DceRpc5AlterContextResp
                        resp = self.sr1(
                            DceRpc5AlterContext(
                                context_elem=self._get_bind_context(interface)
                            ),
                            auth_verifier=CommonAuthVerifier(
                                auth_type=self.ssp.auth_type,
                                auth_level=self.auth_level,
                                auth_context_id=self.session.auth_context_id,
                                auth_value=token,
                            ),
                        )
                        if respcls not in resp:
                            status = GSS_S_FAILURE
                            break
                        if resp.auth_verifier is None:
                            status = GSS_S_COMPLETE
                            break
                        self.sspcontext, token, status = self.ssp.GSS_Init_sec_context(
                            self.sspcontext,
                            input_token=resp.auth_verifier.auth_value,
                            target_name="host/" + self.host,
                        )
            else:
                log_runtime.error("GSS_Init_sec_context failed with %s !" % status)

        # Check context acceptance
        if (
            status == GSS_S_COMPLETE
            and respcls in resp
            and self._check_bind_context(interface, resp.results)
        ):
            self.call_id = 0  # reset call id
            port = resp.sec_addr.port_spec.decode()
            ndr = self.session.ndr64 and "NDR64" or "NDR32"
            self.ndr64 = self.session.ndr64
            if self.verb:
                print(
                    conf.color_theme.success(
                        f"<< {respcls.__name__} port '{port}' using {ndr}"
                    )
                )
            self.session.sspcontext = self.sspcontext
            self._first_time_on_interface = True
            return True
        else:
            if self.verb:
                if DceRpc5BindNak in resp:
                    err_msg = resp.sprintf(
                        "reject_reason: %DceRpc5BindNak.provider_reject_reason%"
                    )
                    print(conf.color_theme.fail("! Bind_nak (%s)" % err_msg))
                    if DceRpc5BindNak in resp:
                        if resp[DceRpc5BindNak].payload and not isinstance(
                            resp[DceRpc5BindNak].payload, conf.raw_layer
                        ):
                            resp[DceRpc5BindNak].payload.show()
                elif DceRpc5Fault in resp:
                    if getattr(resp, "status", 0) != 0:
                        if resp.status in _DCE_RPC_ERROR_CODES:
                            print(
                                conf.color_theme.fail(
                                    f"! {_DCE_RPC_ERROR_CODES[resp.status]}"
                                )
                            )
                        elif resp.status in STATUS_ERREF:
                            print(
                                conf.color_theme.fail(f"! {STATUS_ERREF[resp.status]}")
                            )
                        else:
                            print(conf.color_theme.fail("! Failure"))
                            resp.show()
                    if DceRpc5Fault in resp:
                        if resp[DceRpc5Fault].payload and not isinstance(
                            resp[DceRpc5Fault].payload, conf.raw_layer
                        ):
                            resp[DceRpc5Fault].payload.show()
                else:
                    print(conf.color_theme.fail("! Failure"))
                    resp.show()
            return False

    def bind(self, interface: Union[DceRpcInterface, ComInterface]) -> bool:
        """
        Bind the client to an interface

        :param interface: the DceRpcInterface object
        """
        return self._bind(interface, DceRpc5Bind, DceRpc5BindAck)

    def alter_context(self, interface: Union[DceRpcInterface, ComInterface]) -> bool:
        """
        Alter context: post-bind context negotiation

        :param interface: the DceRpcInterface object
        """
        return self._bind(interface, DceRpc5AlterContext, DceRpc5AlterContextResp)

    def bind_or_alter(self, interface: Union[DceRpcInterface, ComInterface]) -> bool:
        """
        Bind the client to an interface or alter the context if already bound

        :param interface: the DceRpcInterface object
        """
        if not self.session.rpc_bind_interface:
            # No interface is bound
            return self.bind(interface)
        elif self.session.rpc_bind_interface != interface:
            # An interface is already bound
            return self.alter_context(interface)
        return True

    def open_smbpipe(self, name: str):
        """
        Open a certain filehandle with the SMB automaton.

        :param name: the name of the pipe
        """
        self.ipc_tid = self.smbrpcsock.tree_connect("IPC$")
        self.smbrpcsock.open_pipe(name)

    def close_smbpipe(self):
        """
        Close the previously opened pipe
        """
        self.smbrpcsock.set_TID(self.ipc_tid)
        self.smbrpcsock.close_pipe()
        self.smbrpcsock.tree_disconnect()

    def connect_and_bind(
        self,
        host: str,
        interface: DceRpcInterface,
        port: Optional[int] = None,
        timeout: int = 5,
        smb_kwargs={},
    ):
        """
        Asks the Endpoint Mapper what address to use to connect to the interface,
        then uses connect() followed by a bind()

        :param host: the host to connect to
        :param interface: the DceRpcInterface object
        :param port: (optional, NCACN_NP only) the port to connect to
        :param timeout: (optional) the connection timeout (default 5)
        """
        # Connect to the interface using the endpoint mapper
        self.connect(
            host=host,
            interface=interface,
            port=port,
            timeout=timeout,
            smb_kwargs=smb_kwargs,
        )

        # Bind in RPC
        self.bind(interface)

    def epm_map(self, interface):
        """
        Calls ept_map (the EndPoint Manager)
        """
        if self.ndr64:
            ndr_uuid = "NDR64"
            ndr_version = 1
        else:
            ndr_uuid = "NDR 2.0"
            ndr_version = 2
        pkt = self.sr1_req(
            ept_map_Request(
                obj=NDRPointer(
                    referent_id=1,
                    value=UUID(
                        Data1=0,
                        Data2=0,
                        Data3=0,
                        Data4=None,
                    ),
                ),
                map_tower=NDRPointer(
                    referent_id=2,
                    value=twr_p_t(
                        tower_octet_string=bytes(
                            protocol_tower_t(
                                floors=[
                                    prot_and_addr_t(
                                        lhs_length=19,
                                        protocol_identifier=0xD,
                                        uuid=interface.uuid,
                                        version=interface.major_version,
                                        rhs_length=2,
                                        rhs=interface.minor_version,
                                    ),
                                    prot_and_addr_t(
                                        lhs_length=19,
                                        protocol_identifier=0xD,
                                        uuid=ndr_uuid,
                                        version=ndr_version,
                                        rhs_length=2,
                                        rhs=0,
                                    ),
                                    prot_and_addr_t(
                                        lhs_length=1,
                                        protocol_identifier="RPC connection-oriented protocol",  # noqa: E501
                                        rhs_length=2,
                                        rhs=0,
                                    ),
                                    {
                                        DCERPC_Transport.NCACN_IP_TCP: (
                                            prot_and_addr_t(
                                                lhs_length=1,
                                                protocol_identifier="NCACN_IP_TCP",
                                                rhs_length=2,
                                                rhs=135,
                                            )
                                        ),
                                        DCERPC_Transport.NCACN_NP: (
                                            prot_and_addr_t(
                                                lhs_length=1,
                                                protocol_identifier="NCACN_NP",
                                                rhs_length=2,
                                                rhs=b"0\x00",
                                            )
                                        ),
                                    }[self.transport],
                                    {
                                        DCERPC_Transport.NCACN_IP_TCP: (
                                            prot_and_addr_t(
                                                lhs_length=1,
                                                protocol_identifier="IP",
                                                rhs_length=4,
                                                rhs="0.0.0.0",
                                            )
                                        ),
                                        DCERPC_Transport.NCACN_NP: (
                                            prot_and_addr_t(
                                                lhs_length=1,
                                                protocol_identifier="NCACN_NB",
                                                rhs_length=10,
                                                rhs=b"127.0.0.1\x00",
                                            )
                                        ),
                                    }[self.transport],
                                ],
                            )
                        ),
                    ),
                ),
                entry_handle=NDRContextHandle(
                    attributes=0,
                    uuid=b"\x00" * 16,
                ),
                max_towers=500,
                ndr64=self.ndr64,
                ndrendian=self.ndrendian,
            )
        )
        if pkt and ept_map_Response in pkt:
            status = pkt[ept_map_Response].status
            # [MS-RPCE] sect 2.2.1.2.5
            if status == 0x00000000:
                towers = [
                    protocol_tower_t(x.value.tower_octet_string)
                    for x in pkt[ept_map_Response].ITowers.value[0].value
                ]
                # Let's do some checks to know we know what we're doing
                endpoints = []
                for t in towers:
                    if t.floors[0].uuid != interface.uuid:
                        if self.verb:
                            print(
                                conf.color_theme.fail(
                                    "! Server answered with a different interface."
                                )
                            )
                        raise ValueError
                    if t.floors[1].sprintf("%uuid%") != ndr_uuid:
                        if self.verb:
                            print(
                                conf.color_theme.fail(
                                    "! Server answered with a different NDR version."
                                )
                            )
                        raise ValueError
                    if self.transport == DCERPC_Transport.NCACN_IP_TCP:
                        endpoints.append((t.floors[4].rhs, t.floors[3].rhs))
                    elif self.transport == DCERPC_Transport.NCACN_NP:
                        endpoints.append(t.floors[3].rhs.rstrip(b"\x00").decode())
                return endpoints
            elif status == 0x16C9A0D6:
                if self.verb:
                    pkt.show()
                    print(
                        conf.color_theme.fail(
                            "! Server errored: 'There are no elements that satisfy"
                            " the specified search criteria'."
                        )
                    )
                raise ValueError
        print(conf.color_theme.fail("! Failure."))
        if pkt:
            pkt.show()
        raise ValueError("EPM Map failed")


def get_endpoint(
    ip,
    interface,
    transport=DCERPC_Transport.NCACN_IP_TCP,
    ndrendian="little",
    verb=True,
    ssp=None,
    smb_kwargs={},
):
    """
    Call the endpoint mapper on a remote IP to find an interface

    :param ip:
    :param interface:
    :param mode:
    :param verb:
    :param ssp:

    :return: a list of connection tuples for this interface
    """
    client = DCERPC_Client(
        transport,
        # EPM only works with NDR32
        ndr64=False,
        ndrendian=ndrendian,
        verb=verb,
        ssp=ssp,
    )

    if transport == DCERPC_Transport.NCACN_IP_TCP:
        endpoint = 135
    elif transport == DCERPC_Transport.NCACN_NP:
        endpoint = "epmapper"
    else:
        raise ValueError("Unknown transport value !")

    client.connect(ip, endpoint=endpoint, smb_kwargs=smb_kwargs)

    client.bind(find_dcerpc_interface("ept"))
    endpoints = client.epm_map(interface)

    client.close()
    return endpoints



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Tabea Spahn <tabea.spahn@e-mundo.de>

# scapy.contrib.status = skip

from scapy.contrib.automotive.xcp.utils import get_ag, get_max_cto, \
    XCPEndiannessField, StrVarLenField
from scapy.fields import ByteEnumField, ByteField, ShortField, StrLenField, \
    IntField, ThreeBytesField, FlagsField, ConditionalField, XByteField, \
    XIntField, FieldLenField
from scapy.packet import Packet, bind_layers


# ##### CTO COMMANDS ######

# STANDARD COMMANDS

class Connect(Packet):
    commands = {0x00: "NORMAL", 0x01: "USER_DEFINED"}
    fields_desc = [
        ByteEnumField("connection_mode", 0, commands),
    ]


class Disconnect(Packet):
    # DISCONNECT has no data
    pass


class GetStatus(Packet):
    # GET_STATUS has no data
    pass


class Synch(Packet):
    # SYNCH has no data
    pass


class GetCommModeInfo(Packet):
    # GET_COMM_MODE_INFO has no data
    pass


class GetId(Packet):
    """Get identification from slave"""
    types = {0x00: "ASCII",
             0x01: "file_name_without_path_and_extension",
             0x02: "file_name_with_path_and_extension",
             0x03: "URL",
             0x04: "File"
             }
    fields_desc = [ByteEnumField("identification_type", 0x00, types)]


class SetRequest(Packet):
    """Request to save to non-volatile memory"""
    fields_desc = [
        FlagsField("mode", 0, 8, [
            "store_cal_req", "store_daq_req", "clear_daq_req", "x3", "x4",
            "x5", "x6", "x7"]),
        XCPEndiannessField(ShortField("session_configuration_id", 0x00))
    ]


class GetSeed(Packet):
    # Get seed for unlocking a protected resource
    seed_mode = {0x00: "first", 0x01: "remaining"}
    res = {0x00: "resource", 0x01: "ignore"}
    fields_desc = [
        ByteEnumField("mode", 0, seed_mode),
        ByteEnumField("resource", 0, res)
    ]


class Unlock(Packet):
    # Send key for unlocking a protected resource
    fields_desc = [
        FieldLenField("len", None, length_of="seed", fmt="B"),
        StrVarLenField("seed", b"", length_from=lambda p: p.len,
                       max_length=lambda: get_max_cto() - 2)
    ]


class SetMta(Packet):
    # Set Memory Transfer Address in slave
    fields_desc = [
        # specification says: position 1,2 type byte (not WORD) The example(
        # Part 5 Example Communication Sequences ) shows 2 bytes for
        # "reserved"
        # http://read.pudn.com/downloads192/doc/comm/903802/XCP%20-Part%205-%20Example%20Communication%20Sequences%20-1.0.pdf # noqa: E501
        # --> 2 bytes
        XCPEndiannessField(ShortField("reserved", 0)),
        ByteField("address_extension", 0),
        XCPEndiannessField(XIntField("address", 0))
    ]


class Upload(Packet):
    # Upload from slave to master
    fields_desc = [ByteField("nr_of_data_elements", 0)]


class ShortUpload(Packet):
    # Upload from slave to master (short version)
    fields_desc = [
        ByteField("nr_of_data_elements", 0),
        ByteField("reserved", 0),
        XByteField("address_extension", 0),
        XCPEndiannessField(IntField("address", 0))
    ]


class BuildChecksum(Packet):
    # Build checksum over memory range
    fields_desc = [
        # specification says: position 1-3 type byte The example(Part 5
        # Example Communication Sequences ) shows 3 bytes for "reserved"
        # http://read.pudn.com/downloads192/doc/comm/903802/XCP%20-Part%205-%20Example%20Communication%20Sequences%20-1.0.pdf # noqa: E501
        # --> 3 bytes
        XCPEndiannessField(ThreeBytesField("reserved", 0)),
        XCPEndiannessField(XIntField("block_size", 0))
    ]


class TransportLayerCmd(Packet):
    # Refer to transport layer specific command
    sub_commands = {
        0xFF: "GET_SLAVE_ID",
        0xFE: "GET_DAQ_ID",
        0xFD: "SET_DAQ_ID",
    }
    fields_desc = [
        ByteEnumField("sub_command_code", 0xFF, sub_commands),
    ]


class TransportLayerCmdGetSlaveId(Packet):
    echo_mode = {
        0x00: "identify_by_echo",
        0x01: "confirm_by_inverse_echo",
    }

    fields_desc = [
        XByteField("x", 0x58),  # ASCII = X
        XByteField("c", 0x43),  # ASCII = C
        XByteField("p", 0x50),  # ASCII = P
        ByteEnumField("mode", 0x00, echo_mode),
    ]


bind_layers(TransportLayerCmd, TransportLayerCmdGetSlaveId,
            sub_command_code=0xFF)


class TransportLayerCmdGetDAQId(Packet):
    fields_desc = [
        XCPEndiannessField(ShortField("daq_list_number", 0)),
    ]


bind_layers(TransportLayerCmd, TransportLayerCmdGetDAQId,
            sub_command_code=0xFE)


class TransportLayerCmdSetDAQId(Packet):
    sub_command = {
        0xFD: "SET_DAQ_ID",
    }
    fields_desc = [
        XCPEndiannessField(ShortField("daq_list_number", 0)),
        XCPEndiannessField(IntField("can_identifier", 0))
    ]


bind_layers(TransportLayerCmd, TransportLayerCmdSetDAQId,
            sub_command_code=0xFD)


class UserCmd(Packet):
    # Refer to user defined command
    fields_desc = [
        ByteField("sub_command_code", 0),
    ]


# Calibration Commands

class Download(Packet):
    # Download from master to slave
    fields_desc = [
        ByteField("nr_of_data_elements", 0),
        ConditionalField(
            StrLenField("alignment", b"",
                        length_from=lambda pkt: get_ag() - 2),
            lambda pkt: get_ag() > 2),
        StrLenField("data_elements", b"",
                    length_from=lambda pkt: get_max_cto() - 2 if get_ag() == 1
                    else get_max_cto() - get_ag()),
    ]


class DownloadNext(Download):
    # Used for the download from master to slave in block mode
    # Same as "Download", but with different command code
    pass


class DownloadMax(Packet):
    # Download from master to slave (fixed size)
    fields_desc = [
        ConditionalField(
            StrLenField("alignment", b"", length_from=lambda _: get_ag() - 1),
            lambda _: get_ag() > 1),
        StrLenField("data_elements", b"",
                    length_from=lambda _: get_max_cto() - (get_ag() * 2 - 1))
    ]


class ShortDownload(Packet):
    # Download from master to slave (short version)
    fields_desc = [
        FieldLenField("len", None, length_of="data_elements", fmt="B"),
        ByteField("reserved", 0),
        ByteField("address_extension", 0),
        XCPEndiannessField(IntField("address", 0)),
        StrVarLenField("data_elements", b"", length_from=lambda p: p.len,
                       max_length=lambda: get_max_cto() - 8)
    ]


class ModifyBits(Packet):
    # Modify bits
    fields_desc = [
        ByteField("shift_value", 0),
        XCPEndiannessField(ShortField("and_mask", 0)),
        XCPEndiannessField(ShortField("xor_mask", 0))
    ]


# Page Switching commands
class SetCalPage(Packet):
    """Set calibration page"""
    fields_desc = [
        FlagsField("mode", 0, 8,
                   ["ecu", "xcp", "x2", "x3", "x4", "x5", "x6", "all"]),
        ByteField("data_segment_num", 0),
        ByteField("data_page_num", 0)
    ]


class GetCalPage(Packet):
    """Get calibration page"""
    fields_desc = [
        ByteField("access_mode", 0),
        ByteField("data_segment_num", 0)
    ]


class GetPagProcessorInfo(Packet):
    """Get general information on PAG processor"""
    pass


class GetSegmentInfo(Packet):
    """Get specific information for a SEGMENT"""
    info_mode = {
        0x00: "get_basic_address_info",
        0x01: "get_standard_info",
        0x02: "get_address_mapping_info"
    }

    fields_desc = [
        ByteEnumField("mode", 0x00, info_mode),
        ByteField("segment_number", 0),
        ByteField("segment_info", 0),
        ByteField("mapping_index", 0)

    ]


class GetPageInfo(Packet):
    """Get specific information for a PAGE"""
    fields_desc = [
        ByteField("reserved", 0),
        ByteField("segment_number", 0),
        ByteField("page_number", 0)
    ]


class SetSegmentMode(Packet):
    """Set mode for a SEGMENT"""
    fields_desc = [
        FlagsField("mode", 0, 8,
                   ["freeze", "x1", "x2", "x3", "x4", "x5", "x6", "x7"]),
        ByteField("segment_number", 0)
    ]


class GetSegmentMode(Packet):
    """Get mode for a SEGMENT"""
    fields_desc = [
        ByteField("reserved", 0),
        ByteField("segment_number", 0)
    ]


class CopyCalPage(Packet):
    """This command forces the slave to copy one calibration page to another.
    This command is only available if more than one calibration page is defined
    """
    fields_desc = [
        ByteField("segment_num_src", 0),
        ByteField("page_num_src", 0),
        ByteField("segment_num_dst", 0),
        ByteField("page_num_dst", 0)
    ]


class SetDaqPtr(Packet):
    """Data acquisition and stimulation, static, mandatory"""
    fields_desc = [
        ByteField("reserved", 0),
        XCPEndiannessField(ShortField("daq_list_num", 0)),
        ByteField("odt_num", 0),
        ByteField("odt_entry_num", 0)
    ]


class WriteDaq(Packet):
    """Data acquisition and stimulation, static, mandatory"""
    fields_desc = [
        ByteField("bit_offset", 0),
        ByteField("size_of_daq_element", 0),
        ByteField("address_extension", 0),
        XCPEndiannessField(IntField("address", 0))
    ]


class SetDaqListMode(Packet):
    """Set mode for DAQ list"""
    fields_desc = [
        FlagsField("mode", 0, 8,
                   ["x0", "direction", "x2", "x3", "timestamp", "pid_off",
                    "x6", "x7"]),
        XCPEndiannessField(ShortField("daq_list_num", 0)),
        XCPEndiannessField(ShortField("event_channel_num", 0)),
        ByteField("transmission_rate_prescaler", 0),
        ByteField("daq_list_prio", 0)
    ]


class GetDaqListMode(Packet):
    """Get mode from DAQ list"""
    fields_desc = [
        ByteField("reserved", 0),
        XCPEndiannessField(ShortField("daq_list_number", 0))
    ]


class StartStopDaqList(Packet):
    """Start/stop/select DAQ list"""
    mode_enum = {0x00: "stop", 0x01: "start", 0x02: "select"}
    fields_desc = [
        ByteEnumField("mode", 0, mode_enum),
        XCPEndiannessField(ShortField("daq_list_number", 0))
    ]


class StartStopSynch(Packet):
    """Start/stop DAQ lists (synchronously)"""
    mode_enum = {0x00: "stop", 0x01: "start", 0x02: "select"}
    fields_desc = [
        ByteEnumField("mode", 0x00, mode_enum)
    ]


class ReadDaq(Packet):
    """Read element from ODT entry"""
    pass


class GetDaqClock(Packet):
    """Get DAQ clock from slave"""
    pass


class GetDaqProcessorInfo(Packet):
    """Get general information on DAQ processor"""
    pass


class GetDaqResolutionInfo(Packet):
    """Get general information on DAQ processing resolutioin"""
    pass


class GetDaqListInfo(Packet):
    """Get specific information for a DAQ list"""
    fields_desc = [
        ByteField("reserved", 0),
        XCPEndiannessField(ShortField("daq_list_num", 0))
    ]


class GetDaqEventInfo(Packet):
    """Get specific information for an event channel"""
    fields_desc = [
        ByteField("reserved", 0),
        XCPEndiannessField(ShortField("event_channel_num", 0))
    ]

    # Cyclic data transfer - static configuration commands


class ClearDaqList(Packet):
    """Clear DAQ list configuration"""
    fields_desc = [
        ByteField("reserved", 0),
        XCPEndiannessField(ShortField("daq_list_num", 0))
    ]


# Cyclic Data transfer - dynamic configuration commands


class FreeDaq(Packet):
    """Clear dynamic DAQ configuration"""
    pass


class AllocDaq(Packet):
    """Allocate DAQ lists"""
    fields_desc = [
        ByteField("reserved", 0),
        XCPEndiannessField(ShortField("daq_count", 0))
    ]


class AllocOdt(Packet):
    """Allocate ODTs to a DAQ list"""
    fields_desc = [
        ByteField("reserved", 0),
        XCPEndiannessField(ShortField("daq_list_num", 0)),
        ByteField("odt_count", 0)
    ]


class AllocOdtEntry(Packet):
    """Allocate ODT entries to an ODT"""
    fields_desc = [
        ByteField("reserved", 0),
        XCPEndiannessField(ShortField("daq_list_num", 0)),
        ByteField("odt_num", 0),
        ByteField("odt_entries_count", 0)
    ]


# Flash Programming commands

class ProgramStart(Packet):
    """Indicate the beginning of a programming sequence"""
    pass


class ProgramClear(Packet):
    """Clear a part of non-volatile memory"""
    access_mode = {0x00: "absolute_access", 0x01: "functional_access"}
    fields_desc = [
        ByteEnumField("mode", 0, access_mode),
        XCPEndiannessField(ShortField("reserved", 0)),
        XCPEndiannessField(IntField("clear_range", 0))
    ]


class Program(Download):
    """Program a non-volatile memory segment"""
    # Same structure as "Download", but with different command code
    pass


class ProgramReset(Packet):
    """Indicate the end of a programming sequence"""
    pass


class GetPgmProcessorInfo(Packet):
    """Get general information on PGM processor"""
    pass


class GetSectorInfo(Packet):
    """Get specific information for a SECTOR"""
    address_mode = {0x00: "get_address", 0x01: "get_length"}
    fields_desc = [
        ByteEnumField("mode", 0, address_mode),
        ByteField("sector_number", 0)
    ]


class ProgramPrepare(Packet):
    """Prepare non-volatile memory programming"""
    fields_desc = [
        ByteField("not_used", 0),
        XCPEndiannessField(ShortField("code_size", 0))
    ]


class ProgramFormat(Packet):
    """Set data format before programming"""
    fields_desc = [
        ByteField("compression_method", 0),
        ByteField("encryption_mode", 0),
        ByteField("programming_method", 0),
        ByteField("access_method", 0)
    ]


class ProgramNext(Download):
    """Program a non-volatile memory segment (Block Mode)"""
    # Same structure as "Download", but with different command code
    pass


class ProgramMax(DownloadMax):
    """Program a non-volatile memory segment (fixed size)"""
    # Same as "DownloadMax", but with different command code
    pass


class ProgramVerify(Packet):
    """Program  Verify"""
    start_mode = {
        0x00: "request_to_start_internal_routine",
        0x01: "sending_verification_value"
    }
    fields_desc = [
        ByteEnumField("verification_mode", 0, start_mode),
        XCPEndiannessField(ShortField("verification_type", 0)),
        XCPEndiannessField(IntField("verification_value", 0))
    ]



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information

# scapy.contrib.description = Multiprotocol Label Switching (MPLS)
# scapy.contrib.status = loads

from scapy.packet import Packet, bind_layers, Padding
from scapy.fields import (
    BitField,
    ByteField,
    ByteEnumField,
    PacketListField,
    ShortField,
)

from scapy.layers.inet import (
    _ICMP_classnums,
    ICMPExtension_Object,
    IP,
    UDP,
)
from scapy.layers.inet6 import IPv6
from scapy.layers.l2 import Ether, GRE

from scapy.contrib.bier import BIER


class EoMCW(Packet):
    name = "EoMCW"
    fields_desc = [BitField("zero", 0, 4),
                   BitField("reserved", 0, 12),
                   ShortField("seq", 0)]

    def guess_payload_class(self, payload):
        if len(payload) >= 1:
            return Ether
        return Padding


class MPLS(Packet):
    name = "MPLS"
    fields_desc = [BitField("label", 3, 20),
                   BitField("cos", 0, 3),
                   BitField("s", 1, 1),
                   ByteField("ttl", 0)]

    def guess_payload_class(self, payload):
        if len(payload) >= 1:
            if not self.s:
                return MPLS
            ip_version = (payload[0] >> 4) & 0xF
            if ip_version == 4:
                return IP
            elif ip_version == 5:
                return BIER
            elif ip_version == 6:
                return IPv6
            else:
                if payload[0] == 0 and payload[1] == 0:
                    return EoMCW
                else:
                    return Ether
        return Padding


# ICMP Extension

class ICMPExtension_MPLS(ICMPExtension_Object):
    name = "ICMP Extension Object - MPLS (RFC4950)"

    fields_desc = [
        ShortField("len", None),
        ByteEnumField("classnum", 1, _ICMP_classnums),
        ByteField("classtype", 1),
        PacketListField("stack", [], MPLS, length_from=lambda pkt: pkt.len - 4),
    ]


# Bindings

bind_layers(Ether, MPLS, type=0x8847)
bind_layers(IP, MPLS, proto=137)
bind_layers(IPv6, MPLS, nh=137)
bind_layers(UDP, MPLS, dport=6635)
bind_layers(GRE, MPLS, proto=0x8847)
bind_layers(MPLS, MPLS, s=0)
bind_layers(MPLS, IP, label=0)  # IPv4 Explicit NULL
bind_layers(MPLS, IPv6, label=2)  # IPv6 Explicit NULL
bind_layers(MPLS, EoMCW)
bind_layers(EoMCW, Ether, zero=0, reserved=0)
